import features
from flask import Flask, request, jsonify, json
from features.materialize import materialize
from features.custom_schema import custom_schema

schema_fields = [
    {"name": "timestamp", "type": "integer"},
    {"name": "name", "type": "string"},
    {"name": "email", "type": "string"},
    {"name": "deviceid", "type": "integer"}
]

custom_schema = custom_schema(schema_fields)

app = Flask(__name__)

@app.route('/aggregate', methods=['POST'])
def aggregate():
    mongo_connection = {
        'uri': "mongodb+srv://admin:6lHqq9LqgwDlninK@cluster0.aqtcyq2.mongodb.net/?retryWrites=true&w=majority",
        'database': 'kafkastream',
        'rawdata': 'rawdata',
        'collection': 'sparkaggregate'
    }

    try:
        # Define the dynamic Spark SQL queries to be executed
        sparksql = [
            "SELECT COUNT(email) AS email_count_last_24_hours, COUNT(name) AS name_count_last_24_hours, deviceid FROM temp_table WHERE deviceid={deviceid} AND timestamp_unix > {timestamp} - 86400 and timestamp_unix < {timestamp_base} GROUP BY deviceid",
            "SELECT COUNT(email) AS email_count_last_6_days, COUNT(name) AS name_count_last_6_days, deviceid FROM temp_table WHERE deviceid={deviceid} AND timestamp_unix > {timestamp} - 604800 and timestamp_unix < {timestamp_base} GROUP BY deviceid",
            "SELECT COUNT(email) AS email_count_last_7_days, COUNT(name) AS name_count_last_7_days, deviceid FROM temp_table WHERE deviceid={deviceid} AND timestamp_unix > {timestamp} - 604800 and timestamp_unix < {timestamp_base} GROUP BY deviceid",
            "SELECT COUNT(email) AS email_count_last_15_days, COUNT(name) AS name_count_last_15_days, deviceid FROM temp_table WHERE deviceid={deviceid} AND timestamp_unix > {timestamp} - 1296000 and timestamp_unix < {timestamp_base} GROUP BY deviceid",
            "SELECT COUNT(email) AS email_count_last_45_days, COUNT(name) AS name_count_last_45_days, deviceid FROM temp_table WHERE deviceid={deviceid} AND timestamp_unix > {timestamp} - 3.888e+6 and timestamp_unix < {timestamp_base} GROUP BY deviceid",
        ]
        json_data = request.get_json()
        timestamp = json_data.get('timestamp')
        deviceid = json_data.get('deviceid')
        token = json_data.get('token')

        materialize_instance = materialize(mongo_connection, timestamp, "deviceid", deviceid, token, custom_schema.schema, sparksql, partition_column="deviceid")

        aggregated_data_str = materialize_instance.read_data_and_aggregate()
        aggregated_data = json.loads(aggregated_data_str)

        return jsonify(aggregated_data)

    except Exception as e:
        return jsonify({"Result":"No data found","error": str(e)})

if __name__ == '__main__':
    app.run(debug=True, port=5001)



from pyspark.sql.types import IntegerType, StringType, StructField, StructType
class custom_schema:
    def __init__(self, schema_fields):
        self.schema = self.generate_schema(schema_fields)

    def generate_schema(self, schema_fields):
        struct_fields = []

        type_mapping = {
            "integer": IntegerType,
            "string": StringType
        }

        for field in schema_fields:
            field_name = field.get("name")
            field_type = field.get("type")
            field_nullable = field.get("nullable", True)

            if field_name and field_type in type_mapping:
                struct_field = StructField(
                    field_name, type_mapping[field_type](), field_nullable
                )
                struct_fields.append(struct_field)

        return StructType(struct_fields)



import time

import features
from features.create_stream import create_stream
from features.authenticate import authenticate

def authenticate_user():
    sdk = authenticate()
    access_token, token_type = sdk.authenticate('burak', 'burak')
    return access_token

def stream(token):
    kafka_connection = {
        "bootstrap.servers": "buraks-air.lan:9092",
        "group.id": "your-group-id",
        'enable.auto.commit': True,  # Enable automatic commit of offsets
        'auto.offset.reset': 'earliest',
        # Start consuming messages from the beginning of the topic if no offset is stored
        'session.timeout.ms': 6000,  # Timeout for session management
    }

    # Optional: Provide custom MongoDB connection settings
    mongo_connection = {
        'uri': "mongodb+srv://admin:6lHqq9LqgwDlninK@cluster0.aqtcyq2.mongodb.net/?retryWrites=true&w=majority",
        'database': 'kafkastream',
        'rawcollection': 'rawdata',
        'collection': 'sparkaggregate'
    }

    # Create a new instance of the create_stream class
    featurizerai = create_stream(kafka_connection, mongo_connection, "fake-data_test5", "timestamp")

    # Start the Kafka consumer in a separate thread
    featurizerai.start(token)

    featurizerai.stop()

    # # Run the consumer for 60 seconds
    # time.sleep(60)
    #
    # # Stop the consumer and wait for the thread to finish
    # featurizerai.stop()
def main():
    token = authenticate_user()
    stream(token)

if __name__ == "__main__":
    main()



import json
import logging
from datetime import datetime, timedelta
import jwt
from pymongo import MongoClient
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, count
from pyspark.sql.types import StringType, StructType, StructField, TimestampType, IntegerType
from pyspark.sql import functions as F
import certifi
from pymongo.server_api import ServerApi


class materialize:
    def __init__(self, mongo_connection, aggregation_date, groupby_attr, groupby_value, token, schema, sparksql, partition_column=None):
        self.mongo_connection = mongo_connection
        self.aggregation_date = datetime.strptime(aggregation_date, "%Y-%m-%d %H:%M:%S")
        self.groupby = groupby_value
        self.groupby_attr = groupby_attr
        self.token = token
        self.schema = schema
        self.sparksql = sparksql
        self.partition_column = partition_column

        self.spark = SparkSession.builder \
            .appName("IncrementalAggregation") \
            .master("local[*]") \
            .getOrCreate()
        self.spark.sparkContext.setLogLevel("ERROR")

    def read_data_and_aggregate(self):


        # authenticate token
        mongo_client = MongoClient(self.mongo_connection['uri'], server_api=ServerApi('1'), tlsCAFile=certifi.where())
        mongo_db = mongo_client[self.mongo_connection['database']]
        mongo_db_featurizer = mongo_client['featurizer']
        users_collection = mongo_db_featurizer["users"]
        SECRET_KEY = "features"  # Change this to your desired secret key
        TOKEN_EXPIRATION = 3600  # Token expiration in seconds (1 hour)

        try:
            payload = jwt.decode(self.token, SECRET_KEY, algorithms=["HS256"])
            user_id: str = payload.get("user_id")
            if user_id is None:
                print("User not found")
                return ("Invalid token")
            print("Authenticated user: " + user_id)
            user = users_collection.find_one({"userid": user_id})
            if user is None:
                print("User not found")
                return ("Invalid token")
        except Exception as e:
            print(e)
            return ("Invalid token")

        raw_data_collection = mongo_db[self.mongo_connection['rawdata']]
        aggregated_data_collection = mongo_db[self.mongo_connection['collection']]

        raw_data = raw_data_collection.find()
        raw_data_list = [doc for doc in raw_data]
        df = self.spark.createDataFrame(raw_data_list, self.schema)

        # Apply partitioning if the partition_column is specified
        if self.partition_column:
            print("Partitioning the DataFrame based on the column: " + self.partition_column)
            raw_data_collection.create_index(self.partition_column)
            raw_data_collection.create_index(self.groupby_attr)
            df = df.repartition(col(self.partition_column))

        # Filter the DataFrame based on the given device_id
        df = df.filter(col(self.groupby_attr) == self.groupby)
        start_time = self.aggregation_date
        df = df.filter((col(self.groupby_attr) == self.groupby) & (col("timestamp") < start_time.timestamp()))
        df = df.withColumn("timestamp_unix", F.col("timestamp").cast("bigint"))
        df.createOrReplaceTempView("temp_table")

        # Execute the dynamic Spark SQL
        for sql in self.sparksql:
            sql_base = sql.replace("{"+ self.groupby_attr + "}", str(self.groupby))
            hour = int(sql.split("_")[3])
            sql_with_timestamp = sql_base.replace("{timestamp}", str(int((start_time - timedelta(hours=hour)).timestamp())))
            sql_with_timestamp_base = sql_with_timestamp.replace("{timestamp_base}", str(int((start_time).timestamp())))
            c_count = self.spark.sql(sql_with_timestamp_base)
            joined_df = df.join(c_count, self.groupby_attr, "left") \
                .select(df.columns + [col for col in c_count.columns if col not in df.columns])
            df = joined_df

        df = df.drop("timestamp_unix", "timestamp", "name", "email")
        # Convert the aggregated data to JSON

        if df.isEmpty():
            json_aggregated_data = "{'error': 'No data found'}"
        else:
            json_aggregated_data = df.toJSON().collect()[0]

        if json_aggregated_data is not None:
            aggregated_data_collection.update_one(
                {"date": start_time, "deviceid": self.groupby},
                {"$set": {"aggregated_data": json_aggregated_data}},
                upsert=True
            )
        else:
            logging.error("Serialized aggregated data is None, skipping update")

        return json_aggregated_data





import requests
import base64

class authenticate:
    def __init__(self, base_url = 'http://0.0.0.0:4000'):
        self.base_url = base_url

    def authenticate(self, username, password):
        url = f'{self.base_url}/authenticate'
        credentials = f'{username}:{password}'.encode('ascii')
        encoded_credentials = base64.b64encode(credentials).decode('ascii')
        headers = {'Authorization': f'Basic {encoded_credentials}'}
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            data = response.json()
            return data.get('access_token'), data.get('token_type')
        else:
            return None, None



import json
import ast
import logging
from datetime import datetime
from http.client import HTTPException

from motor.motor_asyncio import AsyncIOMotorClient
from pymongo.server_api import ServerApi
from aiokafka import AIOKafkaConsumer
import certifi
import os
import asyncio
import jwt


class create_stream:
    def __init__(self, kafka_connection, mongo_connection=None, topicname="", timestamp_attr="timestamp"):
        self.topicname = topicname
        self.kafka_connection = kafka_connection
        self.mongo_connection = mongo_connection
        self.timestamp_attr = timestamp_attr
        self.stop_event = asyncio.Event()

    def is_epoch(self, timestamp):
        try:
            datetime.fromtimestamp(int(timestamp))
            return True
        except ValueError:
            return False

    async def _run(self, token):
        mongo_client = AsyncIOMotorClient(self.mongo_connection['uri'], server_api=ServerApi('1'),
                                          tlsCAFile=certifi.where())
        mongo_db = mongo_client[self.mongo_connection['database']]
        raw_data_collection = mongo_db[self.mongo_connection['rawcollection']]

        consumer = AIOKafkaConsumer(
            self.topicname,
            bootstrap_servers=self.kafka_connection["bootstrap.servers"],
            group_id=self.kafka_connection["group.id"],
            auto_offset_reset=self.kafka_connection["auto.offset.reset"],
            enable_auto_commit=self.kafka_connection['enable.auto.commit'],
            value_deserializer=lambda m: json.loads(m.decode('utf-8'))
        )

        await consumer.start()
        try:
            while not self.stop_event.is_set():
                msg = await consumer.getone()
                message_value = msg.value
                print(message_value)

                if self.timestamp_attr not in message_value:
                    print(
                        f"Error: Message does not contain the required timestamp attribute '{self.timestamp_attr}'. Skipping message.")
                    continue

                if self.is_epoch(message_value.get(self.timestamp_attr)):
                    if self.is_epoch(message_value.get(self.timestamp_attr)):
                        message_value[self.timestamp_attr] = int(message_value[self.timestamp_attr])
                    else:
                        message_value[self.timestamp_attr] = datetime.now().timestamp()

                await raw_data_collection.insert_one(message_value)
                print("Message received: {}".format(message_value))
        finally:
            await consumer.stop()

    async def start(self, token):
        if not hasattr(self, '_consumer_task') or self._consumer_task.done():
            self.stop_event.clear()
            self._consumer_task = asyncio.create_task(self._run(token))
            print('featurizerai: Started consumer task.')
        else:
            print("Task is already running. Cannot start it again.")

    async def stop(self):
        if hasattr(self, '_consumer_task') and not self._consumer_task.done():
            self.stop_event.set()
            await self._consumer_task
            print('featurizerai: Stopped consumer task.')
        else:
            print("Task is not running. Cannot stop it.")
